/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.flexiscale;

import java.rmi.RemoteException;
import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import com.flexiant.extility.FilterJobList;
import com.flexiant.extility.FlexiScaleSoapBindingStub;
import com.flexiant.extility.Job;


public class FlexiscaleJobs {
	static private final Logger logger = Logger.getLogger(FlexiscaleJobs.class);	
	private FlexiscaleProvider provider = null;
	private FlexiScaleSoapBindingStub service = null;

	FlexiscaleJobs( FlexiscaleProvider provider) throws CloudException{
		this.provider = provider;
		FlexiscaleStub locator = new FlexiscaleStub (this.provider);
		service = locator.getFlexiscaleStub();
	}
	
	/**
	 * @return Iterable<Job>
	 * @throws CloudException
	 * @throws InternalException
	 */
	public Job getJob(String jobId) throws CloudException,InternalException {
		try {
			Job job = service.getJob(Long.valueOf(jobId));
			return job;
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}
	public void cancelJob(String jobId) throws CloudException,InternalException{
		try {
			int returnCode= service.cancelJob(Long.valueOf(jobId));
			if(returnCode !=1 ) throw new CloudException("Failed to delete firewall without explanation.");		
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}	
	}
	
	public boolean isJobRunning(String jobId) throws CloudException,InternalException{
		try {
			int returnCode= service.isJobRunning(Long.valueOf(jobId));
			if(returnCode ==1 ) return true;
			return false;
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
		
	}
	public Iterable<Job> listJobs() throws CloudException,InternalException {
		ArrayList<Job> list = new ArrayList<Job>();
		try {
			Job[] jobs = service.listJobs();
			for(Job job: jobs){
				list.add(job);
			}
			return list;
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}
	/**
	 * 
	 * @param from : minimum Job ID
	 * @param to: maximum Job ID
	 * @param orderBy
	 * @param direction
	 * @throws CloudException
	 * @throws InternalException
	 */
	public Iterable<Job> listJobs(String from, String to, JobOrderBy orderBy, JobOrderDirection direction) throws CloudException,InternalException {
		ArrayList<Job> list = new ArrayList<Job>();
		try {
			FilterJobList joblist = service.filterJobs(Long.valueOf(from), Long.valueOf(to), orderBy.name(), direction.name());
			if(joblist == null) return null;
			Job[] jobs = joblist.getJob_list();
			for(Job job: jobs){
				list.add(job);
			}
			return list;
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}
	public enum JobOrderBy {
		job_id,
		type_id, 
		status,
		started,
		finished,
		description,
		parent_job;
	}
	public enum JobOrderDirection {
		DESC, 
		ASC;
	}
}
