/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.flexiscale.compute;

import java.rmi.RemoteException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;

import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.compute.Volume;
import org.dasein.cloud.compute.VolumeState;
import org.dasein.cloud.compute.VolumeSupport;
import org.dasein.cloud.flexiscale.compute.FlexiscaleDiskProduct;
import org.dasein.cloud.flexiscale.FlexiscaleProvider;
import org.dasein.cloud.flexiscale.FlexiscaleStub;
import org.dasein.cloud.identity.ServiceAction;

import com.flexiant.extility.Disk;
import com.flexiant.extility.FlexiScaleSoapBindingStub;
import com.flexiant.extility.ProductComponent;
import com.flexiant.extility.ProductOffer;
import com.flexiant.extility.Server;
import com.flexiant.extility.VDC;

public class FlexiscaleDisk  implements VolumeSupport {
	
	static private final Logger logger = Logger.getLogger(FlexiscaleDisk.class);
	private FlexiscaleProvider provider = null;
	private FlexiScaleSoapBindingStub service = null;	
	static private List<FlexiscaleDiskProduct> diskProducts;
    static int DiskProductOffer = 3;

    private void loadDiskProduct() {
        Thread t = new Thread() {
            public void run() {
                try { Thread.sleep(600000L); }
                catch( InterruptedException e ) { }
                while( true ) {
                    try {
                        load();
                    }
                    catch( Throwable ignore ) {
                        logger.warn("Error loading Flexiscale products in Volume Support");
                    }

                }
            }
        };         
        t.setName("Flexiscale Product Loader in Volume Support");
        t.setDaemon(true);
        t.start();
        try { load(); }
        catch( Throwable ignore ){ 
        	logger.warn("Error loading Flexiscale disk products in Volume Support");
        }
    }    
    private void load() throws InternalException,CloudException { 	
  
        // Initialization of disk products
        ArrayList<FlexiscaleDiskProduct> disksizes = new ArrayList<FlexiscaleDiskProduct>();
        FlexiscaleDiskProduct diskProduct; 
		ProductOffer[] diskOffers = null;
		try {
			diskOffers = service.listProductOffers(DiskProductOffer);
			for(ProductOffer diskOffer: diskOffers){		
				diskProduct = new FlexiscaleDiskProduct();
				diskProduct.setProductId(diskOffer.getProduct_offer_id());
				diskProduct.setDescription(diskOffer.getDescription());
				ProductComponent[] productComponents = diskOffer.getComponents();	
				String name;
				for(ProductComponent productComponent:productComponents){
					name = productComponent.getName();
					if(name.toUpperCase().contains("DISK")){
						double diskInGb =   productComponent.getQuantity();
						diskProduct.setDiskSizeInGb((int)diskInGb);	
						diskProduct.setName(productComponent.getName());					
					}
				}
				disksizes.add(diskProduct);
			}
	        diskProducts = Collections.unmodifiableList(disksizes);         
		} catch (RemoteException e) {
			logger.error("Error loading Flexiscale disk products in Volume Support ->  " +  e.getMessage());
			throw new CloudException(e);	
		}	
    }
   
	
	FlexiscaleDisk(FlexiscaleProvider provider) throws CloudException{
		this.provider = provider;
		FlexiscaleStub locator = new FlexiscaleStub (this.provider);
		service = locator.getFlexiscaleStub(provider);
		loadDiskProduct();
	}
		
	/**
	 *Flexiscale disk starts with /dev/hda
	 *@param diskID: specify the unique long value of disk id
	 *
	 **/
	@Override
	public void attach(String diskID, String toServerId, String toDevice) throws InternalException, CloudException {
		if(diskID == null || toServerId == null || Long.valueOf(toServerId) == 0L)
			return;
		//The volume can not be attached if the server is running,only when the server is paused
		if(this.checkServerReady(toServerId)){
			int index = getDeviceIndex(toDevice);
			attach(diskID,toServerId,index);
		}else{
			throw new CloudException("The server need to be paused when attaching the disk");
		}
	}
	
	public void attach(String diskID, String toServerId, int index) throws InternalException, CloudException {
		try {
			int returncode = service.attachDisk(Long.parseLong(toServerId), Long.parseLong(diskID), index);
			if(returncode != 1){
				throw  new CloudException( "Attaching dick was not sunccessful");
			}
		} catch (RemoteException e) {
			throw new CloudException(e);
		}
	}
	/**
	 * Get disk product information
	 * @param diskSize: disksize in GB
	 * @return FlexiscaleDiskProduct
	 */
	public FlexiscaleDiskProduct getDiskProduct(int diskSize) throws InternalException, CloudException {
		//Iterable<FlexiscaleDiskProduct> diskSizes = this.listDiskProducts();
		if(diskProducts == null)
			return null;
		//Default is the first one
		FlexiscaleDiskProduct closetProduct = diskProducts.get(0); //
		//Max difference
		int maxDifferenceValue =   Math.abs(diskProducts.get(0).getDiskSizeInGb() - diskSize); 
		for( FlexiscaleDiskProduct product : diskProducts) {
			if( product.getDiskSizeInGb() == diskSize) {
				return product;
			}else{
				// Hardcoe the diskProduct if the sizeInGb is not within the disk product list
				int differenceInGB = Math.abs(product.getDiskSizeInGb()-diskSize );
				if(differenceInGB < maxDifferenceValue){
					closetProduct = product;
				}				
			}
		}
		return closetProduct;		
	}
	/**
	 * Get disk product information
	 * @param diskProductId: disk product offering Id
	 * @return FlexiscaleDiskProduct
	 */
	public FlexiscaleDiskProduct getDiskProduct(long diskProductId) throws InternalException, CloudException {
		//Iterable<FlexiscaleDiskProduct> diskSizes = this.listDiskProducts();
		for( FlexiscaleDiskProduct product : diskProducts) {
			if( product.getProductId() == diskProductId) {
				return product;
			}
		}
		return null;
	}
	
	/**
	 * Create a volume with empty disk if fromSnapshot is null
	 */
	@Override
	public String create(String fromSnapshot, int sizeInGb, String inZoneId) throws InternalException, CloudException {
		FlexiscaleDiskProduct diskProduct= this.getDiskProduct(sizeInGb);
		String inVdcId = this.provider.getRegionIdFromDataCenterId(inZoneId);
		//Hardcode the volume name as enstratus - 2009-09-22 16:47:08
	    SimpleDateFormat sdfDate = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");//dd/MM/yyyy
	    Date now = new Date();
	    String strDate = sdfDate.format(now);
		String volumeName = "enstratus-" + strDate;
		return create(fromSnapshot, diskProduct, volumeName, inVdcId);
	}	
	/**
	 * Create a volume from a image 
	 * @param imageId: unique image Id in flexiscale
	 * @param diskProduct: disk product
	 * @param volumeName: volume name will be created
	 * @param vdcId: virtual data center Id for the volumed to be created
	 * @return volume name
	 * @throws InternalException
	 * @throws CloudException
	 */
	public String create(String imageId, FlexiscaleDiskProduct diskProduct, String volumeName, String vdcId) throws InternalException, CloudException {
		try {
			long flexiscaleImageId = -1l;
			if(imageId != null){
				flexiscaleImageId = Long.parseLong(imageId);
			}
			if(volumeName == null||volumeName.isEmpty()){				
				throw new CloudException("Flexiscale requires to specify a name for the volume to be created");
			}			
			Disk fdisk = service.createDisk(Long.parseLong(vdcId),volumeName ,diskProduct.getProductId(),flexiscaleImageId);
			if(fdisk != null){				
				return String.valueOf(fdisk.getDisk_id());
			}else{
				throw new CloudException("Fail to create disk without explaination");
			}
		} catch (NumberFormatException e) {
			logger.equals(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.equals(e);
			throw new CloudException(e);
		}			
	}	

	/**
	 *  Create a volume from a uploaded pre-existing disk image 
	 * @param url: Specified Required URL for the Image
	 * @param checksum: (Optional) Specify the MD5 Checksum
	 * @param connUser: (Optional) Specify the authenticated user required to access the image
	 * @param connPass: (Optional) Specify the password required to access the image
	 * @param volumeName: volume name
	 * @param diskProduct: Specify the Required Disk ProductOffer (Size)
	 * @param vdcId: Specify the Required VDC ID of which VDC it will be located in
	 * @return
	 * @throws InternalException
	 * @throws CloudException
	 */
	public AsynchronousTask<String> createVolumeFromUpload(String imageUrl, String checksum, 
			String connUser, String connPass, String volumeName, 
			FlexiscaleDiskProduct diskProduct, String vdcId) throws InternalException, CloudException {
		if(imageUrl == null || diskProduct == null || volumeName == null || vdcId == null)
			return null;
		final AsynchronousTask<String> task = new AsynchronousTask<String>();
	    final long fvdcId = Long.parseLong(vdcId);
	    final long diskProductId = diskProduct.getProductId();
	    final String fvolumeName = volumeName;
	    final String fchecksum = checksum;
	    final String fconnUser = connUser;
	    final String fconnPass = connPass;
	    final String fimageUrl = imageUrl;	  
	  
	    Thread t = new Thread() { 
	      public void run() {
	    	  try {
	    		  task.completeWithResult(waitForCreateVolumeFromUpload(fimageUrl, fchecksum, fconnUser, fconnPass, fvolumeName, diskProductId,fvdcId, false, null,task));
	    	  }catch( Throwable t ) {
	    		  task.complete(t);
	    	  }
	      	}
	      };
	      t.setName("Creating volume " + volumeName + "from" + imageUrl);
	      t.setDaemon(true);
	      t.start();
	      return task;			
	}
	
	/**
	 * Implement creating a volume from a uploaded pre-existing disk image 
	 * @param Name: volume name when creating volume only, image name when makeImage is true
	 * @param diskProductofferId:Specify the Required Disk ProductOffer (Size)
	 * @param makeImage: Specify if it should automatically be converted into an ImageTemplate upon creation (0,1)
	 * @param imageDefaultUsername
	 * @param task
	 * @return
	 * @throws CloudException
	 * @throws InternalException
	 */
	public String waitForCreateVolumeFromUpload(String url, String checksum, 
			String connUser, String connPass, String name, long diskProductofferId,
			long vdcId, boolean makeImage, String imageDefaultUsername,
			AsynchronousTask<String> task )throws CloudException, InternalException {
		try {
			long jobId = service.fetchDisk(url, checksum, connUser, connPass, name, diskProductofferId, vdcId, makeImage, imageDefaultUsername);
			if (jobId != -1L){
				return name;
			}
			return null;
		} catch (NumberFormatException e) {
			logger.error(e.getStackTrace());
			throw new InternalException("The imageId for flexiscale should be a long type number. Details -> "+ e);
		} catch (RemoteException e) {
			logger.error(e.getStackTrace());
			throw new CloudException(e);
		}
	}

	@Override
	public void detach(String diskId) throws InternalException, CloudException {
		long fromServerId = getServerFromDiskId(diskId);
		if(fromServerId != 0L){			
			//The volume can not be deattached if the server is running,only when the server is paused
			if(this.checkServerReady(String.valueOf(fromServerId))){
				detach(fromServerId,diskId);
			}else{
				throw new CloudException("The server need to be paused when detaching the disk");
			}
		}else{
			throw new CloudException("The volume is not attached to a server");
		}	
	}
	private void detach( long fromServerId,String diskId) throws InternalException, CloudException {
		try {
			int returncode = service.detachDisk(fromServerId, Long.parseLong(diskId));
			if(returncode != 1){
				throw  new CloudException( "detaching dick was not sunccessful");
			}
		} catch (NumberFormatException e) {
			logger.equals(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}
	private int getDeviceIndex(String device){
		char lastChar = device.charAt(device.length()-1);
		int value = lastChar - 'b';
		switch(value){
			case 0: return 1;
			case 1: return 2;
			case 2: return 3;
			case 3: return 4;
			case 4: return 5;
			default: return -1;		
		}
	}
	
	@Override
	public String getProviderTermForVolume(Locale locale) {
		return "Disk";
	}
	
	@Override
	public Volume getVolume(String diskId) throws InternalException, CloudException {
		try {
			Disk dsk = service.getDisk(Long.valueOf(diskId));
			Volume volume = toVolume(dsk,false);
			return volume;
		} catch (NumberFormatException e) {
			logger.equals(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.equals(e);
			throw new CloudException(e);
		}
	}
	//The volume can only be handled on paused server
	public boolean checkServerReady(String serverId)throws InternalException, CloudException {
		try {
			Server vm = service.getServer(Long.valueOf(serverId));
			if(vm != null){
				/**
				 * (1, 'Error'),(2, 'Running'),(3, 'Starting'),(4, 'Stopping'),(5, 'Stopped'),(6, 'Rebooting'),(7, 'Installing');
				 */	
				switch(vm.getStatus()){
					case 2: return false;
					case 3: return  false;
					case 4: return false; 
					case 5:	return true; 
					case 6:	return false; 
					case 7:	return false; 
					default: return false;//0?		
				}
			}
					
		} catch (NumberFormatException e) {
			logger.equals(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
		return false;
	}

	public long getServerFromDiskId(String diskId) throws CloudException, InternalException {
		try {
			Disk disk = service.getDisk(Long.parseLong(diskId));
			if(disk != null)
				return disk.getServer_id();		
		} catch (NumberFormatException e) {
			logger.equals(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
		return 0L;
	}
	
	public long getPrimaryDiskIdFromServerId(String vmId) throws CloudException, InternalException {
		if(vmId == null)
			return 0L;
		try {
			long serverId = Long.valueOf(vmId);
			String regionId = this.provider.getRegionId();
			Disk[] disks = service.listDisks(Long.valueOf(regionId));
			for(Disk disk : disks){
				if (disk.getServer_id() == serverId ){
					return disk.getDisk_id();
				}
			}	
		} catch (NumberFormatException e) {
			logger.equals(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
		return 0L;
	}
	@Override
	public boolean isSubscribed() throws CloudException, InternalException {
		return true;
	}

	@Override
	public Iterable<String> listPossibleDeviceIds(Platform platform)throws InternalException, CloudException {
        ArrayList<String> list = new ArrayList<String>();        
        if( platform.isWindows() ) {
            list.add("xvdb");
            list.add("xvdc");
            list.add("xvdd");
            list.add("xvde");
            list.add("xvdf");
        }
        else {
            list.add("/dev/hdb");
            list.add("/dev/hdc");
            list.add("/dev/hdd");
            list.add("/dev/hde");
            list.add("/dev/hdf");
        }
        return list;
	}
	/**
	 * List disk product offers provided by flexiscale
	 * @return disk products
	 * @throws InternalException
	 * @throws CloudException
	 */

	public Iterable<FlexiscaleDiskProduct> listDiskProducts()throws InternalException, CloudException {
        ArrayList<FlexiscaleDiskProduct> disksizes = new ArrayList<FlexiscaleDiskProduct>();
        FlexiscaleDiskProduct diskProduct; 
		ProductOffer[] diskOffers = null;
		try {
			diskOffers = service.listProductOffers(DiskProductOffer);
			for(ProductOffer diskOffer: diskOffers){		
				diskProduct = new FlexiscaleDiskProduct();
				diskProduct.setProductId(diskOffer.getProduct_offer_id());
				diskProduct.setDescription(diskOffer.getDescription());
				ProductComponent[] productComponents = diskOffer.getComponents();	
				String name;
				for(ProductComponent productComponent:productComponents){
					name = productComponent.getName();
					if(name.toUpperCase().contains("DISK")){
						double diskInGb =   productComponent.getQuantity();
						diskProduct.setDiskSizeInGb((int)diskInGb);	
						diskProduct.setName(productComponent.getName());					
					}
				}
				disksizes.add(diskProduct);
			}
	       return Collections.unmodifiableList(disksizes);         
		} catch (RemoteException e) {
			logger.error("Error loading Flexiscale disk products in Volume Support ->  " +  e.getMessage());
			throw new CloudException(e);	
		}
	}
		
	/** 
	 *  List disks from all virtual data centers
	 * */
	@Override
	public Iterable<Volume> listVolumes() throws InternalException, CloudException {
		ArrayList<Volume> list = new ArrayList<Volume>();
		try {
			VDC[] vdcs = service.listVDCs();
			for (VDC vdc: vdcs){
				Disk[] dsks = service.listDisks(vdc.getVdc_id());
				for(Disk dsk : dsks){
					Volume volume = toVolume(dsk,false);
					if( volume != null ) {
						list.add(volume);
            		}
				}
			}
			return list;
		} catch (RemoteException e) {
			throw new CloudException(e);
		}
	}

	@Override
	public void remove(String diskId) throws InternalException, CloudException {
		try {
			long fromServerId = getServerFromDiskId(diskId);
			//The volume can be removed only when it is not attached to a sever
			if(fromServerId == 0L ){
				int returncode = service.deleteDisk(Long.valueOf(diskId));
				if(returncode != 1){
					logger.error("Remove disk fails");
					throw new CloudException("Remove disk fails with no explaination");
				}								
			}		

		} catch (NumberFormatException e) {
			throw new InternalException (e);
		} catch (RemoteException e) {
			throw new CloudException(e);
		}
	}
	
	private Volume toVolume(Disk dsk, boolean isCreate) throws CloudException, InternalException{
		Volume volume = new Volume();		
		long ServerId = dsk.getServer_id();
		volume.setProviderVirtualMachineId(String.valueOf(ServerId));
		
		// If just creating the state is set as pending; 
		if(isCreate ){
			volume.setCreationTimestamp(System.currentTimeMillis());
			volume.setCurrentState(VolumeState.PENDING);
		}else{
			volume.setCurrentState(VolumeState.AVAILABLE);
		}		

		volume.setName(dsk.getDisk_name());		
		String volumeRegionId = String.valueOf(dsk.getVdc_id());
		volume.setProviderRegionId(volumeRegionId); 		
		volume.setProviderVirtualMachineId(String.valueOf(dsk.getServer_id()));	
		//volume.setProviderDataCenterId(String.valueOf(dsk.getVdc_id()));
		volume.setProviderDataCenterId(provider.getDataCenterIdFromRegionId(volumeRegionId));
		volume.setProviderVolumeId(String.valueOf(dsk.getDisk_id()));
		volume.setSizeInGigabytes(dsk.getCapacity());		
		
		return volume;		
	}
	@Override
	public String[] mapServiceAction(ServiceAction arg0) {
		// TODO Auto-generated method stub
		return null;
	}
}
