/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

/**
 * @author Qunying Huang, enStratus 
 *	
 *  Flexiscale API has very different implemenation for firewall rules
 *  It will create file wall rule for each ip address while dasein create a security group for all ip address.
 *  To resolve this incomparable issue, will create file wall groups and save the group information in the
 *  custmor metadata. Each firewll group element contains different rule information.
 *  It will assign the rules to a specific address when start a machine associating with this rule. 
 */

/**
 * <securityGroupInfo>
    <item>
       <ownerId>999988887777</ownerId>
       <groupId>sg-455b6c31</groupId>
       <groupName>WebServers</groupName>
       <groupDescription>Web</groupDescription>
       <vpcId/>
       <ipPermissions>
          <item>
             <ipProtocol>tcp</ipProtocol>
             <fromPort>80</fromPort>
             <toPort>80</toPort>
             <groups/>
             <ipRanges>
                <item>
                   <cidrIp>0.0.0.0/0</cidrIp>
                </item>
             </ipRanges>
          </item>
       </ipPermissions>
       <ipPermissionsEgress/>
    	<ipAddressWithinGroup>
        	<item>
        	<ipAddress>
			129.174.65.111
			</ipAddress>
			<firewallId>Id</firewallId>Id
			<fireWallRulesWithinIpAddress>
			<item> rule1 </item>
			<item> rule2 </item>
			<fireWallRulesWithinIpAddress>
			<item>
		</ipAddressWithinGroup> 
		<serverIdWithinGroup>
        	<item>
			serverId1
			<item>
			<item>
			serverId2
			<item>
		</serverIdWithinGroup>       
    </item>
  </securityGroupInfo>
*
*
*This caused a really hugh changes our code design.
*For example, And update this metadata while handling operations, such as launching, terminating VM, associating/releasing IP, authorizing/revoking firewall rules, creating/removing security group.  
*/

package org.dasein.cloud.flexiscale.network;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;


import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.ProviderContext;


import org.dasein.cloud.flexiscale.FlexiscaleProvider;
import org.dasein.cloud.flexiscale.FlexiscaleStub;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.network.Direction;
import org.dasein.cloud.network.Firewall;
import org.dasein.cloud.network.FirewallRule;
import org.dasein.cloud.network.FirewallSupport;

import org.dasein.cloud.network.Permission;
import org.dasein.cloud.network.Protocol;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.flexiant.extility.FlexiScaleSoapBindingStub;

import com.flexiant.extility.VDC;
import com.flexiant.extility.Vlan;

public class FlexiscaleFirewall implements FirewallSupport {

	static private final Logger logger = Logger.getLogger(FlexiscaleIP.class);
	private FlexiscaleProvider provider = null;
	private FlexiScaleSoapBindingStub service = null;
	
	public FlexiscaleFirewall(FlexiscaleProvider provider) throws CloudException {
		this.provider = provider;
		FlexiscaleStub locator = new FlexiscaleStub (this.provider);
		this.service = locator.getFlexiscaleStub();
	} 

	@Override
	public String authorize(@Nonnull String securityGroupId, @Nonnull String cidr, @Nonnull Protocol protocol, int startPort, int endPort) throws CloudException, InternalException {
		if(isRuleExist(securityGroupId,cidr,protocol, startPort,  endPort)){
			logger.debug("Fail to add rule because it exsit!!!");
			return null;
		}		
		Document doc = provider.getDocfromMetaData();
		Element securityGroupItem;
		NodeList blocks = doc.getElementsByTagName(provider.SecurityGroupId__Node_Name);
		if(blocks != null && blocks.getLength() > 0){			
			for( int i=0; i<blocks.getLength(); i++ ) {
				Node node = blocks.item(i);					
	            // Ignore text node with space and comment node
	            if(node.getNodeType() == Node.TEXT_NODE) continue;
	            if(node.getNodeType() == Node.COMMENT_NODE) continue;	  
	            securityGroupItem = (Element)node;
	          
	            if( securityGroupItem.getTextContent().trim().equals(securityGroupId) ) {
	            	if(provider.addAuthorizeInCustomerMetaData((Element)securityGroupItem.getParentNode(), cidr,protocol, startPort,  endPort)){
	            		//Authorization rule for Flexiscale IP address
	            		Collection<String> ipList = provider.getFlexiscaleIpaddressForSecurityGroup(securityGroupId);
	            		if(ipList != null){
	            			for(String ipAddress : ipList){         			
		            		
	            				Long flexiscaleFirewallId = this.getFlexiscaleFirewallIdFromIpAddress(ipAddress);
	            				if(flexiscaleFirewallId == null ){
	            					flexiscaleFirewallId = this.createFileWallInFlexiscale(ipAddress);
	            				}
	            				
	            				String firewallRuleId = authorizeFlexiscaleIpAddress(flexiscaleFirewallId, cidr, protocol, startPort, endPort);
	            				//Add in the metadata
	            				provider.addFirewallRuleForIpAddresseInSecurityGroup(securityGroupId, ipAddress, firewallRuleId);
	            				
		            		}	
	            		}	            		
	            		return securityGroupId;
	            	}
	            }
	          
			}
	           				
		}else{
			throw new CloudException("There is no such security group !!!");
		}
		return null;		
	}
	
	public void authorizeFlexiscaleIpAddressForSecurityGroup(@Nonnull String securityGroupId, String ipAddress, Long flexiscaleFirewallId)throws CloudException, InternalException{
				
/*		//For local debug
 * 		Collection<FirewallRule> listRules = (Collection<FirewallRule>) getRules(securityGroupId);
		
		if(flexiscaleFirewallId == null ){			
			flexiscaleFirewallId = getFlexiscaleFirewallIdFromIpAddress(ipAddress);
			if(flexiscaleFirewallId == null){
				flexiscaleFirewallId = provider.createFlexiscaleFirewall(ipAddress);				
			}		
    	}		
    	for(FirewallRule rule:listRules ){
    		String firewallRuleId = authorizeFlexiscaleIpAddress(flexiscaleFirewallId, rule.getCidr(), rule.getProtocol(), rule.getStartPort(), rule.getEndPort());
    		provider.addFirewallRuleForIpAddresseInSecurityGroup(securityGroupId, ipAddress, firewallRuleId);
    	}*/
		
	   final Long fflexiscaleFirewallId = flexiscaleFirewallId;
	   final String fsecurityGroupId = securityGroupId;
	   final String fipAddress = ipAddress;
		
       Thread t = new Thread() {
            public void run() {
            try {
        		Collection<FirewallRule> listRules = (Collection<FirewallRule>) getRules(fsecurityGroupId);
        		Long flexiscaleFirewallId = fflexiscaleFirewallId;
        		if(flexiscaleFirewallId == null ){			
        			flexiscaleFirewallId = getFlexiscaleFirewallIdFromIpAddress(fipAddress);
        			if(flexiscaleFirewallId == null){
        				flexiscaleFirewallId = provider.createFlexiscaleFirewall(fipAddress);				
        			}		
            	}		
            	for(FirewallRule rule:listRules ){
            		String firewallRuleId = authorizeFlexiscaleIpAddress(flexiscaleFirewallId, rule.getCidr(), rule.getProtocol(), rule.getStartPort(), rule.getEndPort());
            		provider.addFirewallRuleForIpAddresseInSecurityGroup(fsecurityGroupId, fipAddress, firewallRuleId);
            	}
            }catch( Exception e ) {
            	e.printStackTrace();
            }
         }
          
        }; 
        
        t.setName("Revoking rules");
        t.setDaemon(true);
        t.start();		
 
	}
		
	public void revokeFlexiscaleFirewallRulesForSecurityGroup(@Nonnull String securityGroupId, String cidr, Protocol protocol, int startPort, int endPort)throws CloudException, InternalException{
		
		final String  fSecurityGroupId = securityGroupId;
		final String fcidr = cidr;
		final int fstartPort = startPort;
		final int fendPort = endPort;
		final Protocol fprotocol = protocol;
		
    /*  	ArrayList<String> addresses = (ArrayList<String>) provider.getFlexiscaleIpaddressForSecurityGroup(fSecurityGroupId);
    	for(String address :addresses ){
    		Long flexiscaleFirewallId = getFlexiscaleFirewallIdFromIpAddress(address);
    		Collection<Long> inDirectionfirewallRuleIds = getFlexiscaleFirewallRuleId(flexiscaleFirewallId, ConnectionDirection.IN.name(),  fcidr,  fprotocol,  fstartPort,  fendPort);
    		if(inDirectionfirewallRuleIds != null){
    			for(Long firewallRuleId :inDirectionfirewallRuleIds ){
    				revokeFirewallRuleInFlexiscaleIpAddress(firewallRuleId.intValue());
    			}
    			
    		}else{            		
        		//Continue?
    			Collection<Long> outDirectionFirewallRuleIds = getFlexiscaleFirewallRuleId(flexiscaleFirewallId, ConnectionDirection.OUT.name(),  fcidr,  fprotocol,  fstartPort,  fendPort);	
    			if(outDirectionFirewallRuleIds != null){
    				for(Long firewallRuleId :inDirectionfirewallRuleIds ){
    					revokeFirewallRuleInFlexiscaleIpAddress(firewallRuleId.intValue());
        			}
        			
        		}
    		}            		
    	}*/
        Thread t = new Thread() {
            public void run() {
            try {
              	ArrayList<String> addresses = (ArrayList<String>) provider.getFlexiscaleIpaddressForSecurityGroup(fSecurityGroupId);
            	for(String address :addresses ){
            		Long flexiscaleFirewallId = getFlexiscaleFirewallIdFromIpAddress(address);
            		Collection<Long> inDirectionfirewallRuleIds = getFlexiscaleFirewallRuleId(flexiscaleFirewallId, ConnectionDirection.IN.name(),  fcidr,  fprotocol,  fstartPort,  fendPort);
            		if(inDirectionfirewallRuleIds != null){
            			for(Long firewallRuleId :inDirectionfirewallRuleIds ){
            				revokeFirewallRuleInFlexiscaleIpAddress(firewallRuleId.intValue());
            			}
            			
            		}else{            		
                		//Continue?
            			Collection<Long> outDirectionFirewallRuleIds = getFlexiscaleFirewallRuleId(flexiscaleFirewallId, ConnectionDirection.OUT.name(),  fcidr,  fprotocol,  fstartPort,  fendPort);	
            			if(outDirectionFirewallRuleIds != null){
            				for(Long firewallRuleId :inDirectionfirewallRuleIds ){
            					revokeFirewallRuleInFlexiscaleIpAddress(firewallRuleId.intValue());
                			}
                			
                		}
            		}            		
            	}
            }
            catch( Exception e ) {
            	e.printStackTrace();
            }
         }
          
        }; 
        
        t.setName("Revoking rules");
        t.setDaemon(true);
        t.start();		
	}
	
	
	
	private String authorizeFlexiscaleIpAddress( long flexiscaleFirewallId, @Nonnull String cidr, @Nonnull Protocol protocol, int startPort, int endPort) throws CloudException, InternalException {

		String firewallRuleName = "FirewallRule";
		String action = BoundPolicy.ALLOW.name();
	
		//default connection state as "ALL", default ConnectionDirection as IN
		return authorize(flexiscaleFirewallId,cidr,action, firewallRuleName,  ConnectionState.ALL, ConnectionDirection.IN,protocol,startPort, endPort );
	}
	/**
	 * @param, firewallRuleName, firewallId, action,direction, protocol id, sequence No, default mask would be 32 if not specified
	 * sequenceNo is used to specify the sequence of the firewall rule in the database
	 */	
	private String authorize(long flexiscaleFirewallId, String cidr,String action, String firewallRuleName, ConnectionState state, ConnectionDirection direction, Protocol protocol, int startPort, int endPort) throws CloudException, InternalException {
		
		com.flexiant.extility.FirewallRule firewallRule = new com.flexiant.extility.FirewallRule();
	
		String ipAddress="0.0.0.0";
		int mask = 32;
		if(cidr != null){
			String[] ipInfo = cidr.split("/");
			ipAddress = ipInfo[0];
			if(ipInfo.length >1){
				mask = Integer.valueOf(ipInfo[1]);
			}
		}		
		
		firewallRule.setBit_mask(mask);		
		firewallRule.setFirewall_id(flexiscaleFirewallId);
		firewallRule.setAction(action);
		firewallRule.setState(state.name());
		firewallRule.setRule_name(firewallRuleName);
		firewallRule.setRemote_ipaddress(ipAddress);
		firewallRule.setRemote_port(startPort);
		firewallRule.setLocal_port(endPort);
		firewallRule.setDirection(direction.name());
		if(direction.equals(Direction.EGRESS)){			
			firewallRule.setLocal_port(startPort);
			firewallRule.setRemote_port(endPort);
		}
				
		int protocolId = this.getFlexiscaleProtocolFromDaseinProtoal(protocol);
		firewallRule.setFirewall_protocol_id(protocolId);
		
		try {
			
			int sequenceNo = getFirewallRuleNumbers(flexiscaleFirewallId, direction.name());
			firewallRule.setSequence_no(sequenceNo);
			
			com.flexiant.extility.FirewallRule ffirewallRule = this.service.createFirewallRule(firewallRule, sequenceNo);
			if(ffirewallRule == null){
				throw new CloudException("Fail to authorize firewall rule"); 
			}else{
				return String.valueOf(ffirewallRule.getFirewall_rule_id());
			}			 
		} catch (RemoteException e) {
			throw new CloudException(e);
		}
	}
	

	private boolean isRuleExist( @Nonnull String providerFirewallId,  @Nonnull String cidr, @Nonnull Protocol protocol, int startPort, int endPort) throws InternalException, CloudException{
		Collection<FirewallRule> listRules = (Collection<FirewallRule>) getRules(providerFirewallId);
		for(FirewallRule rule: listRules){
			if(rule.getCidr().equals(cidr)
					&& rule.getProtocol().equals(protocol)
					&& rule.getStartPort() == startPort 
					&& rule.getEndPort() == endPort){
				return true;				
			}		
		}		
		return false;
	}

	
	
	public boolean addIpAddressInSecurityGroup(@Nonnull String securityGroupId,String ipAddress) throws InternalException, CloudException {
		
		Document doc = provider.getDocfromMetaData();
		Element securityGroupItem = (Element) provider.getSecurityGroupItem(doc, securityGroupId);
		
		if(securityGroupItem!=null){
			NodeList ipAddressWithinGroupBlocks = securityGroupItem.getElementsByTagName("ipAddressWithinGroup");
		    if(ipAddressWithinGroupBlocks == null || ipAddressWithinGroupBlocks.getLength() == 0){ //Indicate that there is no ip address using this security group rules
		    	Element ipAddressWithinGroupElmt = doc.createElement("ipAddressWithinGroup");
		    	Element ipAddressesItem = doc.createElement("item");
		       
		    	ipAddressesItem.setTextContent(ipAddress);
		    	ipAddressWithinGroupElmt.appendChild(ipAddressWithinGroupElmt);

		    	securityGroupItem.appendChild(ipAddressWithinGroupElmt);
		    	
		    	return provider.setCustomerMetadata(doc);
		    }else{
		    	// Check if the ipAddress already in the security Group
		       // Element ipAddressWithinGroupItem = null;
		        for(int j=0; j<ipAddressWithinGroupBlocks.getLength(); j++ ) {
		        	NodeList targets = ipAddressWithinGroupBlocks.item(j).getChildNodes();			
		        	for( int k=0; k<targets.getLength(); k++ ) {
		        		Node item = targets.item(k);
		        		if( item.getNodeName().equals("item") && item.getTextContent().equals(ipAddress) ) {
		        			logger.error("The ip address already in the security group !!!!");
			        		return false;
		        		}
		        	}	
		        		        	
		        	//If the ip address is not in the address group, then add the address

	    			Element ipAddressItemToBeAdded = doc.createElement("item");
	    			ipAddressItemToBeAdded.setTextContent(ipAddress);
	    			ipAddressWithinGroupBlocks.item(j).appendChild(ipAddressItemToBeAdded);
	    			return provider.setCustomerMetadata(doc);
		        }
		    }
					
		}else{
			throw new CloudException("There is no such security group !!!");
		}
		return false;	
		
	}
  
  

	@Override
	public String create(String name, String description) throws InternalException,CloudException {
		return createInVLAN(name,description, null );		
	}

	public Long getFlexiscaleFirewallIdFromIpAddress(String ipAddress) throws InternalException, CloudException{
		ArrayList<com.flexiant.extility.Firewall> list = (ArrayList<com.flexiant.extility.Firewall>) this.listFlexiscaleFirewall(provider.getRegionId());
		for(com.flexiant.extility.Firewall firewall : list ){
			if(firewall.getIp_address().equals(ipAddress)){
				return firewall.getFirewall_id();				
			}			
		}
		return null;		 
	}
	
	
	public Long createFileWallInFlexiscale(String ipAddress) throws InternalException,CloudException {
		Long flexiscaleFirewallId = this.getFlexiscaleFirewallIdFromIpAddress(ipAddress);
		if(flexiscaleFirewallId != null){
			return flexiscaleFirewallId;			
		}else{
			return createFileWallInFlexiscale(ipAddress, BoundPolicy.ALLOW, BoundPolicy.ALLOW, this.provider.getRegionId());	
		}
		
	}
	
	/**
	 * @param ipAddress: ip address
	 * @param defualtInboundPolicy: in bound policy(Drop,allow, reject)
	 * @param defaultOutboundPolicy: out bound policy
	 * @param vdcId: vi
	 * @return firewall rule id
	 * @throws CloudException 
	 * @throws InternalException 
	 */
	
	public Long createFileWallInFlexiscale(String ipAddress, BoundPolicy defualtInboundPolicy, BoundPolicy defaultOutboundPolicy, String vdcId) throws InternalException,CloudException {
		try {			
			
			com.flexiant.extility.Firewall firewall = service.createFirewall(ipAddress, defualtInboundPolicy.name(), defaultOutboundPolicy.name(), Long.valueOf(vdcId));
			if(firewall != null){
				return firewall.getFirewall_id();
			}else{
				throw new CloudException("Failed to create firewall without explanation.");				
			}
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}
	
	@Override
	public String createInVLAN(@Nonnull String name, @Nonnull String description,  String providerVlanId)throws InternalException, CloudException {
		String metaData = provider.getMetaData();
		Document doc = null;
		
		if(metaData != null && !metaData.equals("")){
			doc = provider.parseStringToDom(metaData);
		}else{
			doc = provider.createDomDocument();
			Element root = doc.createElement(provider.Customer_MetaData_Root);
			doc.appendChild(root);			
		}		
		Document result = provider.buildSecurityGroupInfoNode(doc, name, description, providerVlanId);
		if(provider.setCustomerMetadata(result)){
			return name;
		}else{
			return null;
		}	
	}
	
	public Long createInVLAN(String ipAddress,  BoundPolicy defualtInboundPolicy, BoundPolicy defaultOutboundPolicy, String vLanId)throws InternalException, CloudException {
		long vdcId = this.getVdcFromVlan(Long.valueOf(vLanId));
		if(vdcId != -1)		
			return createFileWallInFlexiscale(ipAddress, defualtInboundPolicy, defaultOutboundPolicy, String.valueOf(vdcId));
		return null;
	}
	
	public Long createFirewallInFlexiscaleInVLAN(String ipAddress,  BoundPolicy defualtInboundPolicy, BoundPolicy defaultOutboundPolicy, String vLanId)throws InternalException, CloudException {
		long vdcId = this.getVdcFromVlan(Long.valueOf(vLanId));
		if(vdcId != -1)		
			return createFileWallInFlexiscale(ipAddress, defualtInboundPolicy, defaultOutboundPolicy, String.valueOf(vdcId));
		return null;
	}
	
	/**
	 * @param firewallId; direction has two option: IN or OUT action
	 * */
	private int getFirewallRuleNumbers(long firewallId, String direction)throws CloudException, InternalException{
		try {
			com.flexiant.extility.FirewallRule[] fFirewallRules = service.listFirewallRules(firewallId, direction);
			if(fFirewallRules != null){
				return fFirewallRules.length;
			}
		} catch (RemoteException e) {
			throw new CloudException(e);
		}
		return 0;		
	}
	
	private Collection<Long> getFlexiscaleFirewallRuleId(long firewallId, String direction,String cidr, Protocol protocol, int startPort, int endPort )throws CloudException, InternalException{
		try {
			ArrayList<Long> list = new ArrayList<Long>();
			com.flexiant.extility.FirewallRule[] fFirewallRules = service.listFirewallRules(firewallId, direction);
			for(com.flexiant.extility.FirewallRule rule : fFirewallRules ){
				FirewallRule dRule = toFirewallRule(rule);
				
				if(!(rule.getDirection().equals(direction)))
					continue;
				if(!dRule.getCidr().equals(cidr))
					continue;
				if(!dRule.getProtocol().equals(protocol))
					continue;
				if(!(dRule.getStartPort()== startPort))
					continue;
				if(!(dRule.getEndPort() == endPort))
					continue;				
				 list.add(rule.getFirewall_rule_id());			
			}
			return list;
			
		} catch (RemoteException e) {
			throw new CloudException(e);
		}
				
	}
	private long getVdcFromVlan( long vLanId) throws InternalException, CloudException {
		try {
			VDC[] vdcs = service.listVDCs();
			if(vdcs == null) return -1l;
			long vdcId = -1l;
			for(VDC vdc: vdcs){
				vdcId= vdc.getVdc_id();
				Vlan[] vlans = service.listVLANs(vdcId);
				if(vlans == null)
					continue;
				for(Vlan vlan: vlans){
					if(vlan.getVlan_id() == vLanId)
						return vdcId;						
				}				
			}
			return vdcId;			
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}	


	public void deleteInFlexiscale(String firewallId) throws InternalException, CloudException {
		try {
			int returnCode = this.service.deleteFirewall(Integer.valueOf(firewallId));
		
			if(returnCode !=1 ) 
				throw new CloudException("Failed to delete firewall without explanation.");
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}
	
	@Override
	public void delete(String securityGroupId) throws InternalException, CloudException {
		provider.deleteInCustomerMetaData(securityGroupId);
	}
		
	/**
	 * Flexiscale does not return the the firewall based on the firewall Id. 
	 * Need to list all firewall first
	 */
	@Override
	public Firewall getFirewall(String firewallId) throws InternalException,CloudException {
		ArrayList<Firewall> firewallLists = (ArrayList<Firewall>) list();
		if(firewallLists != null){
			for(Firewall firewall:firewallLists ){
				if(firewall.getProviderFirewallId().equals(firewallId))
					return firewall;
			}
		}
		return null;
	}

	@Override
	public String getProviderTermForFirewall(Locale locale) {
		return "Firewall";
	}

	/**
	 * @param  direction: Specifies a Direction, Valid Inputs are 'IN' or 'OUT'
	 * */	 	 
	public Collection<FirewallRule> getRules(String firewallId, String direction) throws InternalException, CloudException{
		 ArrayList<FirewallRule> list = new ArrayList<FirewallRule>();
		 try {
			com.flexiant.extility.FirewallRule[] fFirewallRules = this.service.listFirewallRules(Long.valueOf(firewallId), direction);
			for(com.flexiant.extility.FirewallRule fFirewallRule: fFirewallRules){
				FirewallRule dFirewallRule = toFirewallRule(fFirewallRule);
				if(dFirewallRule != null)
					list.add(dFirewallRule);
			}
			return list;			
		 } catch (NumberFormatException e) {
			 logger.error(e);
			 throw new InternalException(e);
		 } catch (RemoteException e) {
			 logger.error(e);
			 throw new CloudException(e);
		 }	 
	 }

	 /**
	* @param  protocalId: Flexicale protocal Id
	* @return dasein-cloud protocol
	* */
	public Protocol getDaseinProtocolFromFlexiscaleProtocolId(int protocalId)
	{
		switch(protocalId){
			case 1: return Protocol.ICMP; 
			case 6: return Protocol.TCP; 
			case 17: return Protocol.UDP;
			default: return null;
		}
	}
	/** 
	* @param  protocol: Dasein-cloud protocol
	* @return Flexicale protocal Id. -1 means any protocol
	* */
	public int getFlexiscaleProtocolFromDaseinProtoal(Protocol protocol)
	{
		switch(protocol){
			case ICMP: return 1;
			case TCP: return 6; 
			case UDP: return 17;
			default: return -1;
		}
	}
	

	@Override
	public Collection<FirewallRule> getRules(String securityGroupId ) throws InternalException, CloudException {
		Node item = provider.getSecurityGroupItem(securityGroupId);		
        return getRulesFromNode(securityGroupId,item);
	}
	
	private Collection<FirewallRule>  getRulesFromNode(String securityGroupId, Node item ){
		ArrayList<FirewallRule> list = new ArrayList<FirewallRule>();
 		
		if( item != null){
			NodeList attrs = item.getChildNodes();
			for( int k=0; k<attrs.getLength(); k++ ) {
				Node attr = attrs.item(k);
				if( attr.getNodeName().equals("ipPermissions") ) {
					NodeList subList = attr.getChildNodes();
					
					for( int l=0; l<subList.getLength(); l++ ) {
						Node sub = subList.item(l);
						
						if( sub.getNodeName().equals("item") ) {
							list.addAll(toFirewallRules(securityGroupId, sub));
						}
					}
				}
	
			}
		}
		return list;	
	}
	
	public Collection<FirewallRule> getRulesInFlexiscale(String firewallId ) throws InternalException, CloudException {
		ArrayList<FirewallRule> list = new ArrayList<FirewallRule>();
		ArrayList<FirewallRule> inFirewallRule = (ArrayList<FirewallRule>) getRules(firewallId, "IN");
		if(inFirewallRule !=null)
			list.addAll(inFirewallRule);
		ArrayList<FirewallRule> outFirewallRule = (ArrayList<FirewallRule>) getRules(firewallId, "OUT");
		if(outFirewallRule !=null)
			list.addAll(outFirewallRule);
		return list;
	}

	@Override
	public Collection<Firewall> list() throws InternalException, CloudException {
		return listFirewallInCustomerMetaData();	
	}
	public Collection<Firewall> listFirewallInCustomerMetaData() throws InternalException, CloudException {
		ArrayList<Firewall> list = new ArrayList<Firewall>();
		String metaData = provider.getMetaData();
		Document doc = provider.parseStringToDom(metaData);		
		if(doc == null || metaData == null || metaData.equals("")){
			throw new CloudException("There is no such firewall Information in the "+ provider.META_DATA_TYPE );
		}else{			
			NodeList blocks = doc.getElementsByTagName("groupName");			
			if(blocks != null && blocks.getLength() > 0 ){	
				for( int i=0; i<blocks.getLength(); i++ ) {
					Node node = blocks.item(i);					
		            // Ignore text node with space and comment node
		            if(node.getNodeType() == Node.TEXT_NODE) continue;
		            if(node.getNodeType() == Node.COMMENT_NODE) continue;	  
		        	Firewall firewall = toFirewall(provider.getContext(), node.getParentNode());
	            	if( firewall != null ) {
	            		list.add(firewall);
	            	}	            	
				}
			}	
		}
		return list;
	}
	
	
	public Collection<Firewall> listFirewallInFlexiscale() throws InternalException, CloudException {
		String regionId = provider.getRegionId();
		return list(regionId);	
	}
	
	
	public Collection<Firewall> list(String vdcId) throws InternalException, CloudException {
		ArrayList<Firewall> list = new ArrayList<Firewall>();
		try {
			com.flexiant.extility.Firewall[] fFirewalls = this.service.listFirewalls(Long.valueOf(vdcId));
			for(com.flexiant.extility.Firewall fFirewall: fFirewalls){
				Firewall dFirewall = toFirewall(fFirewall);
				if(dFirewall != null)
					list.add(dFirewall);
			}
			return list;			
		} catch (NumberFormatException e) {
			throw new InternalException(e);
		} catch (RemoteException e) {
			throw new CloudException(e);
		}	 
	}
	
	public Collection<com.flexiant.extility.Firewall> listFlexiscaleFirewall(String vdcId) throws InternalException, CloudException {
		ArrayList<com.flexiant.extility.Firewall> list = new ArrayList<com.flexiant.extility.Firewall>();
		try {
			com.flexiant.extility.Firewall[] fFirewalls = this.service.listFirewalls(Long.valueOf(vdcId));
			for(com.flexiant.extility.Firewall fFirewall: fFirewalls){
					list.add(fFirewall);
			}
			return list;			
		} catch (NumberFormatException e) {
			throw new InternalException(e);
		} catch (RemoteException e) {
			throw new CloudException(e);
		}	 
	}
	
	private Firewall toFirewall(com.flexiant.extility.Firewall fFirewall) throws InternalException, CloudException{
		if(fFirewall == null) return null;
		Firewall dFirewall = new Firewall();
		dFirewall.setProviderFirewallId(String.valueOf(fFirewall.getFirewall_id()));
		dFirewall.setRegionId(provider.getContext().getRegionId());
		dFirewall.setActive(true);
		dFirewall.setAvailable(true);		
		// There is no name argument in the flexiscale, use ip address instead 
		dFirewall.setName(fFirewall.getIp_address());
		return dFirewall;
	}
	
	private Firewall toFirewall(@Nonnull ProviderContext ctx, @Nullable Node node) throws InternalException, CloudException{
	     if( node == null ) {
	            return null;
	        }
	        String fwName = null, fwId = null, fwDesc = null;
			NodeList attrs = node.getChildNodes();
			Firewall firewall = new Firewall();
	        String regionId = ctx.getRegionId();
			String vpcId = null;

	        if( regionId == null ) {
	            return null;
	        }
	        firewall.setRegionId(regionId);
	        firewall.setAvailable(true);
	        firewall.setActive(true);
			for( int i=0; i<attrs.getLength(); i++ ) {
				if(attrs.item(i).getNodeType() == Node.TEXT_NODE) continue;
		        if(attrs.item(i).getNodeType() == Node.COMMENT_NODE) continue;	
				
		        Element attr = (Element)attrs.item(i);
				String name;
				
				name = attr.getNodeName();

				if( name.equals("groupName") ) {
	                fwName = attr.getFirstChild().getNodeValue().trim();
				}
				else if( name.equals("groupDescription") ) {
					fwDesc = attr.getFirstChild().getNodeValue().trim();
				}
				else if( name.equals("groupId") ) {
	                fwId = attr.getFirstChild().getNodeValue().trim();			    
				}
				else if( name.equals("vpcId") ) {
				    if( attr.hasChildNodes() ) {
				        vpcId = attr.getFirstChild().getNodeValue();
				        if( vpcId != null ) {
				            vpcId = vpcId.trim();
				        }
				    }
				}
			}
	        if( fwId == null ) {
	            if( fwName == null ) {
	                return null;
	            }
	            fwId = fwName;
	        }
			if( fwName == null ) {
			    fwName = fwId;
			}
	        firewall.setProviderFirewallId(fwId);
	        firewall.setName(fwName);
	        if( fwDesc == null ) {
	            fwDesc = fwName;
	        }
	        firewall.setDescription(fwDesc);
			if( vpcId != null ) {
			    firewall.setName(firewall.getName() + " (VPC " + vpcId + ")");
			    firewall.setProviderVlanId(vpcId);
			}			
			return firewall;
	}



	@Override
	public void revoke(String securityGroupId, String cidr, Protocol protocol, int startPort, int endPort) throws CloudException, InternalException {
		//throw new OperationNotSupportedException("Argmments are not matched");
		
		Document doc = provider.getDocfromMetaData();
		Node securityGroupItem = provider.getSecurityGroupItem(doc, securityGroupId);
		
		NodeList attrs = securityGroupItem.getChildNodes();
		
		for( int k=0; k<attrs.getLength(); k++ ) {
			Node attr = attrs.item(k);
			if( attr.getNodeName().equals("ipPermissions") ) {
				NodeList subList = attr.getChildNodes();
				
				for( int l=0; l<subList.getLength(); l++ ) {
					Node sub = subList.item(l);
					
					if( sub.getNodeName().equals("item") ) {
						Collection<FirewallRule> listRules = toFirewallRules(securityGroupId, sub);						
						
						for(FirewallRule rule: listRules){
							if(rule.getCidr().equals(cidr)
									&& rule.getProtocol().equals(protocol)
									&& rule.getStartPort() == startPort 
									&& rule.getEndPort() == endPort){
								//Remove the node
								attr.removeChild(sub);
								//Reset the metadata
								provider.setCustomerMetadata(doc);
								// Revoke the firewll rules in the flexiscale
								revokeFlexiscaleFirewallRulesForSecurityGroup(securityGroupId, cidr, protocol, startPort, endPort);
								
								return;										
							}		
						}		
					}
				}
			}

		}	
		
	}
	
	public void revokeAllRulesWithInSecurityGroupForIpAddress(String securityGroupId, String ipAddress) throws CloudException, InternalException {
		final String fsecurityGroupId = securityGroupId;
		final String fipAddress  = ipAddress;
		Thread t = new Thread() {
            public void run() {
            try {
            	
            	Collection<FirewallRule> listRules = (Collection<FirewallRule>) getRules(fsecurityGroupId);
            	
            	for(FirewallRule rule :listRules ){
            		Long flexiscaleFirewallId = getFlexiscaleFirewallIdFromIpAddress(fipAddress);
            		Collection<Long> inDirectionfirewallRuleIds = getFlexiscaleFirewallRuleId(flexiscaleFirewallId, ConnectionDirection.IN.name(), rule.getCidr(),  rule.getProtocol(),  rule.getStartPort(),  rule.getEndPort());
            		if(inDirectionfirewallRuleIds != null){
            			for(Long firewallRuleId :inDirectionfirewallRuleIds ){
            				revokeFirewallRuleInFlexiscaleIpAddress(firewallRuleId.intValue());
            			}
            			
            		}else{            		
                		//Continue?
            			Collection<Long> outDirectionFirewallRuleIds = getFlexiscaleFirewallRuleId(flexiscaleFirewallId, ConnectionDirection.OUT.name(),  rule.getCidr(),  rule.getProtocol(),  rule.getStartPort(),  rule.getEndPort());
            			if(outDirectionFirewallRuleIds != null){
            				for(Long firewallRuleId :inDirectionfirewallRuleIds ){
            					revokeFirewallRuleInFlexiscaleIpAddress(firewallRuleId.intValue());
                			}
                			
                		}
            		}            
            		
            		
            	}
            }
            catch( Exception e ) {
            	e.printStackTrace();
            }
         }
          
        }; 
        
        t.setName("Revoking rules");
        t.setDaemon(true);
        t.start();		
		
	}
	

	public void revokeAllRulesWithInSecurityGroupForIpAddress(Collection<String> flexiscaleRuleIds) throws CloudException, InternalException {
	
		final Collection<String> fflexiscaleRuleIds =  flexiscaleRuleIds;
		Thread t = new Thread() {
            public void run() {
            try {
            	
            	for(String ruleId :fflexiscaleRuleIds ){
            		revokeFirewallRuleInFlexiscaleIpAddress(Integer.valueOf(ruleId));
            			
            	}
          
            } catch( Exception e ) {
            	e.printStackTrace();
            }
         }
          
        }; 
        
        t.setName("Revoking rules");
        t.setDaemon(true);
        t.start();		
		
		
	}
	
	
	
	
	private void revokeFirewallRuleInFlexiscaleIpAddress(int firewallRuleId) throws CloudException, InternalException {
		try {
			int returnCode = this.service.deleteFirewallRule(firewallRuleId);
			if(returnCode !=1 ) 
				throw new CloudException("Failed to delete firewall Rule without explanation.");
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}
	private FirewallRule toFirewallRule(com.flexiant.extility.FirewallRule fFirewallRule )throws InternalException, CloudException {
		FirewallRule dFirewallRule = new FirewallRule();
		if(fFirewallRule == null) return null;
		dFirewallRule.setCidr(fFirewallRule.getRemote_ipaddress() + "/" + fFirewallRule.getBit_mask());
		
		String direction = fFirewallRule.getDirection();
		
		// Default as INGRESS
		dFirewallRule.setDirection(Direction.INGRESS);	
		dFirewallRule.setStartPort(fFirewallRule.getRemote_port());
		dFirewallRule.setEndPort(fFirewallRule.getLocal_port());
		
		
		
		if(direction != null){
			if(direction.equalsIgnoreCase("OUT")){
				dFirewallRule.setDirection(Direction.EGRESS);
				dFirewallRule.setStartPort(fFirewallRule.getLocal_port());
				dFirewallRule.setEndPort(fFirewallRule.getRemote_port());
			}
		}
		dFirewallRule.setFirewallId(String.valueOf(fFirewallRule.getFirewall_id()));
		 
		// Flexiscale has three actions: allow, reject and drop
		String action = fFirewallRule.getAction().toUpperCase();
		if(action.equalsIgnoreCase("ALLOW")){
			dFirewallRule.setPermission(Permission.ALLOW);
		}else if(action.equalsIgnoreCase("REJECT")){
			dFirewallRule.setPermission(Permission.DENY);
		}		 
		dFirewallRule.setProtocol(getDaseinProtocolFromFlexiscaleProtocolId(fFirewallRule.getFirewall_protocol_id()));
		return dFirewallRule;		 
	}
	
	private @Nonnull Collection<FirewallRule> toFirewallRules(@Nonnull String securityGroupId, @Nullable Node node) {
	    ArrayList<FirewallRule> rules = new ArrayList<FirewallRule>();
        
        if( node == null ) {
            return rules;
        }
	    ArrayList<String> cidrs = new ArrayList<String>();
		NodeList attrs = node.getChildNodes();
        int startPort = -2, endPort = -2;
		Protocol protocol = Protocol.TCP;
		
		for( int i=0; i<attrs.getLength(); i++ ) {
			Node attr = attrs.item(i);
			String name;
			
			name = attr.getNodeName();
			if( name.equals("ipProtocol") ) {
			    String val = attr.getFirstChild().getNodeValue().trim();
			    
			    if( !val.equals("") && !val.equals("-1") ) {
			        protocol = Protocol.valueOf(attr.getFirstChild().getNodeValue().trim().toUpperCase());
			    }
			    else {
			        protocol = Protocol.ICMP;
			    }
			}
			else if( name.equals("fromPort") ) {
				startPort = Integer.parseInt(attr.getFirstChild().getNodeValue().trim());
			}
			else if( name.equals("toPort") ) {
				endPort = Integer.parseInt(attr.getFirstChild().getNodeValue().trim());
			}
			else if( name.equals("ipRanges") ) {
				if( attr.hasChildNodes() ) {
					NodeList children = attr.getChildNodes();
				
					for( int j=0; j<children.getLength(); j++ ) {
						Node child = children.item(j);
					
						if( child.getNodeName().equals("item") ) {
						    if( child.hasChildNodes() ) { 
						        NodeList targets = child.getChildNodes();
						        
						        for( int k=0; k<targets.getLength(); k++ ) {
						            Node cidr = targets.item(k);

						            if( cidr.getNodeName().equals("cidrIp") ) {
						                cidrs.add(cidr.getFirstChild().getNodeValue());
						            }
						        }
						    }
						}
					}
				}
			}
		}
		for( String cidr : cidrs ) {
		    FirewallRule rule = new FirewallRule();

		    rule.setFirewallId(securityGroupId);
		    rule.setCidr(cidr);
		    rule.setEndPort(endPort);
		    rule.setStartPort(startPort);
		    rule.setProtocol(protocol);
		    rules.add(rule);
		}
		return rules;		
	}
	
	@Override
	public boolean isSubscribed() throws CloudException, InternalException {
		// TODO Auto-generated method stub
		return true;
	}
	
	@Override
	public String[] mapServiceAction(ServiceAction arg0) {
		// TODO Auto-generated method stub
		return new String[0];
	}
	
	/**
	 * Bound policy
	 */
	
	public enum BoundPolicy {	
	    ALLOW, 
	    DROP, 
	    REJECT;
	}	
	/**
	 * <p>
	 *  Connection direction for firewall rules.
	 * </p>
	 */
	public enum ConnectionDirection {	
		 /**
	     * packets come in 
	     */	
	    IN, 
	    /**
	     * come out
	     */
	    OUT;
	}
	public enum ConnectionState {
		 /**
	     *Any connections 
	     */	
	    ALL, 
	    /**
	     * NEW Connection
	     */
	    NEW, 
	    /**
	     *Existing connections 
	     */
	    EXISTING;
	}

}
